from argparse import SUPPRESS, ArgumentParser
from multiprocessing import cpu_count
from pathlib import Path

from mc_trimmer.main import CRITERIA_MAPPING, main

from . import Paths
from .__version__ import __version__


def run():
    parser = ArgumentParser(
        prog="mctrimmer",
        description=f"Trim a minecraft dimension based on per-chunk criteria. v{__version__}",
        add_help=False,
    )

    parser.add_argument(
        "-h",
        "--help",
        help="Show this help message and exit.",  # Default implementation is not capitalized
        action="help",
        default=SUPPRESS,
    )
    parser.add_argument(
        "-b",
        "--backup",
        dest="backup_dir",
        help="Backup regions affected by trimming to this directory. Defaults to './backup'",
        nargs="?",
        default=None,
        const="./backup",
    )
    parser.add_argument(
        "-i",
        "--input-region",
        dest="input_dir",
        help="Directory to source the dimension files from. If no output directory is specified, in-place editing will be performed.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output-region",
        dest="output_dir",
        help="Directory to store the dimension files to. If unspecified, in-place editing will be performed by taking the input directory instead.",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "-p",
        "--parallel",
        dest="threads",
        help="Parallelize the task. If no thread count is specified, the number of cpu cores -1 is taken instead.",
        nargs="?",
        type=int,
        default=None,
        const=cpu_count() - 1,
    )

    parser.add_argument(
        "-c",
        "--criteria",
        dest="trimming_criteria",
        choices=[k for k in CRITERIA_MAPPING.keys()],
        help="Pre-defined criteria by which to determmine if a chunk should be trimmed or not.",
        required=True,
    )

    # Parse
    args, _ = parser.parse_known_args()

    inp = Path(args.input_dir)
    outp = Path(args.output_dir) if args.output_dir is not None else inp
    backup = Path(args.backup_dir) if args.backup_dir else None
    threads: int | None = args.threads

    paths = Paths(inp, outp, backup)

    main(threads=threads, paths=paths, trimming_criteria=args.trimming_criteria)



import os
import struct
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Callable, Generic, Iterable, Self, Type, TypeVar


class Paths:
    def __init__(self, inp: Path, outp: Path, backup: Path | None = None) -> None:
        if backup == inp:
            raise Exception("Input and backup directories cannot be the same.")
        if backup == outp:
            raise Exception("Output and backup directories cannot be the same.")
        if not inp.exists():
            raise Exception("Input directory must exist.")

        self.inp_region: Path = inp / "region"
        self.inp_poi: Path = inp / "poi"
        self.inp_entities: Path = inp / "entities"

        self.outp_region: Path = outp / "region"
        self.outp_poi: Path = outp / "poi"
        self.outp_entities: Path = outp / "entities"

        self.backup_region: Path | None = None
        self.backup_poi: Path | None = None
        self.backup_entities: Path | None = None

        if backup is not None:
            self.backup_region = backup / "region"
            self.backup_poi = backup / "poi"
            self.backup_entities = backup / "entities"

            Paths.__assert_writable(self.backup_region)
            Paths.__assert_writable(self.backup_poi)
            Paths.__assert_writable(self.backup_entities)

        self.inp_region.mkdir(exist_ok=True)
        self.inp_poi.mkdir(exist_ok=True)
        self.inp_entities.mkdir(exist_ok=True)

        Paths.__assert_writable(self.outp_region)
        Paths.__assert_writable(self.outp_poi)
        Paths.__assert_writable(self.outp_entities)

    @staticmethod
    def __assert_writable(path: Path) -> None:
        path.mkdir(exist_ok=True, parents=True)
        if not os.access(path, os.W_OK):
            raise Exception(f'Path "{path}" is not writable. Aborting.')


T = TypeVar("T")
S = TypeVar("S", bound="Serializable")


class Strategy(Generic[T]):
    def __init__(self, typ: bytes, payload_size: int, unpack: bytes, resulting_type: Type[T]) -> None:
        self.typ: bytes = typ
        self.payload_size: int = payload_size
        self.unpack: bytes = unpack
        self.resulting_type: Type[T] = resulting_type


BYTE_STRATEGY = Strategy(typ=b"\x01", payload_size=1, unpack=b">c", resulting_type=int)
INT_STRATEGY = Strategy(typ=b"\x03", payload_size=4, unpack=b">i", resulting_type=int)
LONG_STRATEGY = Strategy(typ=b"\x04", payload_size=8, unpack=b">Q", resulting_type=int)
FLOAT_STRATEGY = Strategy(typ=b"\x05", payload_size=4, unpack=b">f", resulting_type=float)
DOUBLE_STRATEGY = Strategy(typ=b"\x06", payload_size=8, unpack=b">d", resulting_type=float)


class Sizes(IntEnum):
    LOCATION_DATA_SIZE = 4 * 1024
    TIMESTAMPS_DATA_SIZE = 4 * 1024
    CHUNK_SIZE_MULTIPLIER = 4 * 1024
    CHUNK_HEADER_SIZE = 4 + 1


class Meta(type):
    def __mul__(mcs: Type[S], i: int) -> Callable[[], "ArrayOfSerializable[S]"]:
        """With T = Type[Serializable], T * int = ArrayofSerializable[T] of size int"""
        assert Serializable in mcs.__mro__

        def curry():
            return ArrayOfSerializable[mcs](mcs, i)

        return curry


class Serializable(metaclass=Meta):
    @abstractmethod
    def __bytes__(self) -> bytes:
        ...

    @classmethod
    @abstractmethod
    def from_bytes(cls, data: bytes) -> Self:
        ...

    @classmethod
    @property
    @abstractmethod
    def SIZE(cls) -> int:
        ...


class ArrayOfSerializable(list[S]):
    def __init__(self, cls: type[S], len: int) -> None:
        self._len: int = len
        self._cls: Type[Serializable] = cls

    def from_bytes(self, data: bytes) -> Self:
        for i in range(self._len):
            obj: S = self._cls.from_bytes(data[i * self._cls.SIZE : (i + 1) * self._cls.SIZE])  # type: ignore
            self.append(obj)
        return self

    def __bytes__(self) -> bytes:
        ret = b""
        for data in self:
            ret += bytes(data)
        return ret


class SerializableLocation(Serializable):
    def __init__(self, offset: int = 0, size: int = 0) -> None:
        self.offset = offset
        self.size = size

    def __repr__(self) -> str:
        return str(f"{self.__class__}: {self.offset}+{self.size}")

    def __lt__(self, other: Self):
        return self.offset < other.offset

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        offset, size = struct.unpack(">IB", b"\x00" + data)  # 3 byte offset, 1 byte size
        return cls(offset=offset, size=size)

    def __bytes__(self) -> bytes:
        return struct.pack(">IB", self.offset, self.size)[1:]

    @classmethod
    @property
    def SIZE(cls) -> int:
        return 4


class Timestamp(Serializable):
    def __init__(self, timestamp: int = 0) -> None:
        self.timestamp = timestamp

    def __repr__(self) -> str:
        return str(f"{self.__class__}: {self.timestamp}")

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        (timestamp,) = struct.unpack(">I", data)
        return cls(timestamp=timestamp)

    def __bytes__(self) -> bytes:
        return struct.pack(">I", self.timestamp)

    @classmethod
    @property
    def SIZE(cls) -> int:
        return 4


LocationData = SerializableLocation * 1024
TimestampData = Timestamp * 1024


@dataclass
class ChunkDataBase(Generic[S]):
    data: S  # Data must be Serializable.
    location: SerializableLocation
    timestamp: Timestamp
    index: int

    def __lt__(self, other: Self) -> bool:
        return self.index < other.index

    def __eq__(self, other: "ChunkDataBase[S]") -> bool:
        return issubclass(self.__class__, other.__class__) and self.index == other.index


class ChunkDataDict(dict[int, ChunkDataBase[S]]):
    def append(self, element: ChunkDataBase[S]):
        self.setdefault(element.index, element)


class RegionLike(ABC):
    @abstractmethod
    def reset_chunk(self, index: int) -> None:
        ...

    @abstractmethod
    def __bytes__(self) -> bytes:
        ...

    def save_to_file(self, file: Path) -> None:
        data = bytes(self)
        if len(data) > Sizes.LOCATION_DATA_SIZE + Sizes.TIMESTAMPS_DATA_SIZE:
            with open(file, "wb") as f:
                f.write(data)
                print(f"Written {file}")
        else:
            print(f"Deleting {file}")
            if file.exists() and file.is_file():
                os.remove(file)

    @staticmethod
    def get_regions(path: str | Path) -> Iterable[str]:
        p: Path = Path(path)
        if p.exists() and p.is_dir():
            return (f.name for f in p.glob("*.mca") if f.is_file())
        raise Exception(f"Invalid input <{p}>")

    @staticmethod
    def to_bytes(data: ChunkDataDict) -> bytes:
        offset: int = 2
        locations: bytearray = bytearray()
        timestamps: bytearray = bytearray()
        chunks: bytearray = bytearray()

        # Adjust offsets and sizes, store chunk data
        for cd in sorted(data.values(), key=lambda combo: combo.location.offset):
            data_bytes: bytes = bytes(cd.data)
            length = len(data_bytes)

            assert length % 4096 == 0
            if length == 0 and cd.location.size != 0:
                pass
            cd.location.size = length // 4096
            if cd.location.size == 0:
                cd.location.offset = 0
                cd.timestamp.timestamp = 0
            else:
                chunks += data_bytes
                cd.location.offset = offset
                offset += cd.location.size

        # Convert tables to binary
        previous = -1
        for cd in sorted(data.values(), key=lambda combo: combo.index):
            if cd.index - previous > 1:
                bytes_to_add = bytearray(b"\x00\x00\x00\x00") * (cd.index - previous - 1)
                locations += bytes_to_add
                timestamps += bytes_to_add
            locations += bytes(cd.location)
            timestamps += bytes(cd.timestamp)
            previous = cd.index
        bytes_to_add = bytearray(b"\x00\x00\x00\x00") * (1024 - previous - 1)
        locations += bytes_to_add
        timestamps += bytes_to_add

        return bytes(locations + timestamps + chunks)


def fast_get_property(decompressed_data: bytes, name: bytes, strategy: Strategy[T]) -> T:
    """Quick-fetch property by seeking through the byte-stream.

    If a property can appear more than once, this will break!"""
    prop_sequence = strategy.typ + struct.pack(">H", len(name)) + name
    start = decompressed_data.find(prop_sequence)
    if start < 0:
        raise Exception(f"Prop '{name.decode()}' not found!")
    (value,) = struct.unpack(
        strategy.unpack,
        decompressed_data[start + len(prop_sequence) : start + len(prop_sequence) + strategy.payload_size],
    )
    return value



import os
import tempfile
from pathlib import Path
from typing import Callable

import pytest

from mc_trimmer import *
from mc_trimmer.entities import Entity

current_dir = Path(os.path.dirname(__file__))
input_dir = current_dir / "in"
output_dir = current_dir / "out"
# test_dir = current_dir / "debug"
# test_dir.mkdir(exist_ok=True)


@pytest.hookimpl(tryfirst=True)
def pytest_exception_interact(call):
    raise call.excinfo.value


@pytest.hookimpl(tryfirst=True)
def pytest_internalerror(excinfo):
    raise excinfo.value


@pytest.mark.parametrize(
    "file,filter",
    [
        ("region/simple.mca", None),
        ("region/remove_one.mca", lambda chunk: chunk.xPos == 1 and chunk.zPos == 288),
        ("region/r.0.0.mca", lambda chunk: chunk.xPos == 0 and chunk.zPos == 0),
        ("region/checkerboard.mca", lambda chunk: (chunk.xPos + chunk.zPos) % 2),
        ("region/complex_checkerboard.mca", lambda chunk: (chunk.xPos + chunk.zPos) % 2),
    ],
)
def test_RegionFile(file: str, filter: Callable[[Chunk], bool] | None):
    input_file = input_dir / file
    output_file = output_dir / file

    region = RegionFile.from_file(input_file)
    if filter is not None:
        region.trim(filter)

    b = bytes(region)

    # region.save_to_file(test_dir / file)

    with open(output_file, "+rb") as f:
        a = f.read()

    t = a == b
    if not t:
        assert False


@pytest.mark.parametrize(
    "file,filter",
    [
        ("entities/simple.mca", None),
        ("entities/remove_one.mca", lambda entity: entity.contains_id("minecraft:chicken")),
    ],
)
def test_EntityFile(file: str, filter: Callable[[Entity], bool] | None):
    input_file = input_dir / file
    output_file = output_dir / file

    entities = EntitiesFile.from_file(input_file)
    if filter is not None:
        entities.trim(filter)

    b = bytes(entities)

    with open(output_file, "+rb") as f:
        a = f.read()

    t = a == b

    if not t:
        assert False


@pytest.mark.parametrize(
    "file,filter",
    [
        ("simple.mca", None),
    ],
)
def test_all(file: str, filter: Callable[[Chunk, Entity], bool]):
    expected_paths = Paths(
        inp=Path(input_dir),
        outp=Path(output_dir),
    )

    with tempfile.TemporaryDirectory() as tmp_dir:
        paths = Paths(
            inp=Path(input_dir),
            outp=Path(tmp_dir),
        )
        manager = RegionManager(paths)
        region: Region = manager.open_file(file)
        if filter is not None:
            manager.trim(region, filter)
        manager.save_to_file(region, file)

        with open(expected_paths.outp_region / file, "rb") as correct, open(paths.outp_region / file, "rb") as got:
            a = correct.read()
            b = got.read()
            t = a == b
            assert t

        with open(expected_paths.outp_entities / file, "rb") as correct, open(paths.outp_entities / file, "rb") as got:
            a = correct.read()
            b = got.read()
            t = a == b
            assert t



import os
import struct
import zlib
from pathlib import Path
from typing import Callable, Self

from mc_trimmer.primitives import (
    INT_STRATEGY,
    LONG_STRATEGY,
    ChunkDataBase,
    ChunkDataDict,
    LocationData,
    RegionLike,
    Serializable,
    Sizes,
    TimestampData,
    fast_get_property,
)

# LOG = logging.getLogger(__name__)


class Chunk(Serializable):
    def __init__(
        self,
        length: int = 0,
        compression: int = 2,
        data: bytes = b"",
        compressed_data: bytes = b"",
    ) -> None:
        self._compression: int = compression
        self._compressed_data: bytes = compressed_data

        if length > 0:
            self.decompressed_data = zlib.decompress(data)[3:]  # 3 bytes removes root tag opening
            pass

    @property
    def InhabitedTime(self) -> int:
        velue = fast_get_property(self.decompressed_data, b"InhabitedTime", LONG_STRATEGY)
        assert velue >= 0
        return velue

    @property
    def xPos(self) -> int:
        return fast_get_property(self.decompressed_data, b"xPos", INT_STRATEGY)

    @property
    def yPos(self) -> int:
        return fast_get_property(self.decompressed_data, b"yPos", INT_STRATEGY)

    @property
    def zPos(self) -> int:
        return fast_get_property(self.decompressed_data, b"zPos", INT_STRATEGY)

    @classmethod
    def from_bytes(cls: type[Self], data: bytes) -> Self:
        length, compression = struct.unpack(">IB", data[: Sizes.CHUNK_HEADER_SIZE])
        nbt_data = data[Sizes.CHUNK_HEADER_SIZE :]  # Sizes.CHUNK_HEADER_SIZE + length - 1]
        assert compression == 2
        post_chunk_data = data[Sizes.CHUNK_HEADER_SIZE + length :]
        if len(post_chunk_data) > 0:
            if post_chunk_data[0] != 0:
                pass
                # print(f"Warning: post-chunk data was padded with non-zero values: {bytes(post_chunk_data[:100])}")
        return cls(length=length, compression=compression, data=nbt_data, compressed_data=data)

    def conditional_reset(self, condition: Callable[[Self], bool]) -> bool:
        if self._compressed_data != b"":
            if condition(self):
                self._compressed_data = b""
                return True
        return False

    def __bytes__(self) -> bytes:
        return bytes(self._compressed_data)

    @property
    def SIZE(self) -> int:
        return len(self._compressed_data)


class RegionFile(RegionLike):
    def __init__(self, chunk_location_data: bytes, timestamps_data: bytes, data: bytes) -> None:
        self.chunk_data: ChunkDataDict[Chunk] = ChunkDataDict[Chunk]()
        self.dirty: bool = False

        locations = LocationData().from_bytes(chunk_location_data)
        timestamps = TimestampData().from_bytes(timestamps_data)

        for i, (loc, ts) in enumerate(zip(locations, timestamps, strict=False)):
            chunk: Chunk
            if loc.size > 0:
                assert loc.offset >= 2
                start = loc.offset * Sizes.CHUNK_SIZE_MULTIPLIER
                data_slice = data[start : start + loc.size * Sizes.CHUNK_SIZE_MULTIPLIER]
                chunk = Chunk.from_bytes(data_slice)

                # Tests:
                # b = bytes(chunk)
                # a = bytes(data_slice)
                # assert a == b
                self.chunk_data.append(ChunkDataBase(data=chunk, location=loc, timestamp=ts, index=i))
        return

    def __bytes__(self) -> bytes:
        return RegionFile.to_bytes(data=self.chunk_data)

    def trim(self, condition: Callable[[Chunk], bool]):
        for i, cd in self.chunk_data.items():
            self.dirty |= cd.data.conditional_reset(condition)

    @classmethod
    def from_file(cls, region: Path) -> Self:
        with open(region, "+rb") as f:
            data = memoryview(f.read()).toreadonly()
            chunk_location_data: bytes = data[: Sizes.LOCATION_DATA_SIZE]
            timestamps_data: bytes = data[
                Sizes.LOCATION_DATA_SIZE : Sizes.LOCATION_DATA_SIZE + Sizes.TIMESTAMPS_DATA_SIZE
            ]
            return RegionFile(chunk_location_data, timestamps_data, data)

    def reset_chunk(self, index: int) -> None:
        popped = self.chunk_data.pop(index, None)
        self.dirty |= popped is not None



import zlib
from .primitives import *


class Entity(Serializable):
    def __init__(
        self,
        length: int = 0,
        compression: int = 2,
        data: bytes = b"",
        compressed_data: bytes = b"",
    ) -> None:
        self._compression: int = compression
        self._compressed_data: bytes = compressed_data
        self.decompressed_data: bytes = b""

        if length > 0:
            self.decompressed_data = zlib.decompress(data)[3:]  # 3 bytes removes root tag opening

    def contains_id(self, id: str) -> bool:
        if len(self.decompressed_data) == 0:
            return False
        bytes_id: bytes = id.encode()
        size: bytes = struct.pack(">H", len(bytes_id))
        sub: bytes = b"\x08\x00\x02id" + size + bytes_id
        return sub in self.decompressed_data

    @classmethod
    def from_bytes(cls: type[Self], data: bytes) -> Self:
        length, compression = struct.unpack(">IB", data[: Sizes.CHUNK_HEADER_SIZE])
        nbt_data = data[Sizes.CHUNK_HEADER_SIZE :]  # Sizes.CHUNK_HEADER_SIZE + length - 1]
        assert compression == 2
        post_chunk_data = data[Sizes.CHUNK_HEADER_SIZE + length :]
        if len(post_chunk_data) > 0:
            if post_chunk_data[0] != 0:
                pass
                # print(f"Warning: post-chunk data was padded with non-zero values: {bytes(post_chunk_data[:100])}")
        return cls(length=length, compression=compression, data=nbt_data, compressed_data=data)

    def __bytes__(self) -> bytes:
        return bytes(self._compressed_data)

    @property
    def SIZE(self) -> int:
        return len(self._compressed_data)


class EntitiesFile(RegionLike):
    def __init__(self, location_data: bytes, timestamp_data: bytes, data: bytes) -> None:
        self.entity_data: ChunkDataDict[Entity] = ChunkDataDict[Entity]()
        self.dirty: bool = False

        if len(location_data) > 0:
            locations: ArrayOfSerializable[SerializableLocation] = LocationData().from_bytes(location_data)
            timestamps: ArrayOfSerializable[Timestamp] = TimestampData().from_bytes(timestamp_data)

            for i, (loc, ts) in enumerate(zip(locations, timestamps, strict=False)):
                if loc.size > 0:
                    assert loc.offset >= 2
                    start = loc.offset * Sizes.CHUNK_SIZE_MULTIPLIER
                    entity_data = data[start : start + loc.size * Sizes.CHUNK_SIZE_MULTIPLIER]

                    d = ChunkDataBase(data=Entity.from_bytes(entity_data), location=loc, timestamp=ts, index=i)
                    self.entity_data.append(d)

    def __bytes__(self) -> bytes:
        return RegionLike.to_bytes(self.entity_data)

    @classmethod
    def from_file(cls, file: Path) -> Self:
        with open(file, "+rb") as f:
            data = memoryview(f.read()).toreadonly()
            chunk_location_data: bytes = data[: Sizes.LOCATION_DATA_SIZE]
            timestamps_data: bytes = data[
                Sizes.LOCATION_DATA_SIZE : Sizes.LOCATION_DATA_SIZE + Sizes.TIMESTAMPS_DATA_SIZE
            ]
            return EntitiesFile(chunk_location_data, timestamps_data, data)

    def trim(self, condition: Callable[[Entity], bool]):
        to_delete: list[int] = []
        for i, cd in self.entity_data.items():
            if condition(cd.data):
                to_delete.append(i)
        for i in to_delete:
            self.reset_chunk(i)

    def reset_chunk(self, index: int) -> None:
        popped = self.entity_data.pop(index, None)
        self.dirty |= popped is not None



import shutil
from dataclasses import dataclass
from functools import partial
import sys
import traceback
from typing import Callable, Iterable

from multiprocess.pool import Pool

from mc_trimmer.entities import EntitiesFile, Entity
from mc_trimmer.primitives import Paths, RegionLike
from mc_trimmer.regions import Chunk, RegionFile


@dataclass
class Region:
    file_name: str
    region: RegionFile
    entities: EntitiesFile

    def trim(self, condition: Callable[[Chunk, Entity], bool]):
        indexes_to_delete: list[int] = []
        for i, cd in self.region.chunk_data.items():
            condition_met: bool = False
            if self.entities is not None:
                ed = self.entities.entity_data.get(i, None)
                if ed is None:
                    condition_met = condition(cd.data, Entity())
                else:
                    condition_met = condition(cd.data, ed.data)
            if condition_met:
                indexes_to_delete.append(i)
        for i in indexes_to_delete:
            self.region.reset_chunk(i)
            if self.entities is not None:
                self.entities.reset_chunk(i)

        pass

    def iterate(self) -> Iterable[tuple[int, Chunk, Entity]]:
        full_outer_join = set(self.region.chunk_data.keys()) | set(self.entities.entity_data.keys())
        for i in full_outer_join:
            c = self.region.chunk_data.get(i, None)
            c = c.data if c is not None else Chunk()

            e = self.entities.entity_data.get(i, None)
            e = e.data if e is not None else Entity()
            yield (i, c, e)

    def reset_chunk(self, index: int):
        self.region.reset_chunk(index)
        self.entities.reset_chunk(index)


class RegionManager:
    def __init__(self, paths: Paths) -> None:
        self._paths: Paths = paths

    def open_file(self, file_name: str) -> Region:
        region = RegionFile.from_file(self._paths.inp_region / file_name)

        if (self._paths.inp_entities / file_name).exists():
            entities = EntitiesFile.from_file(self._paths.inp_entities / file_name)
        else:
            entities = EntitiesFile(b"", b"", b"")

        return Region(region=region, entities=entities, file_name=file_name)

    def trim(self, region: Region, condition: Callable[[Chunk, Entity], bool]) -> None:
        for i, c, e in region.iterate():
            if condition(c, e):
                region.reset_chunk(i)

    def save_to_file(self, region: Region, file_name: str) -> None:
        if region.region.dirty:
            if self._paths.backup_region is not None:
                shutil.copy2(self._paths.inp_region / file_name, self._paths.backup_region / file_name)
            region.region.save_to_file(self._paths.outp_region / file_name)
        else:
            print(f"Region unchanged: {file_name}")
            if self._paths.inp_region != self._paths.outp_region:
                shutil.copy2(self._paths.inp_region / file_name, self._paths.outp_region / file_name)

        if region.entities.dirty:
            if self._paths.backup_entities is not None:
                shutil.copy2(self._paths.inp_entities / file_name, self._paths.backup_entities / file_name)
            region.entities.save_to_file(self._paths.outp_entities / file_name)
        else:
            print(f"Entities unchanged: {file_name}")
            if (
                self._paths.inp_entities != self._paths.outp_entities
                and (self._paths.inp_entities / file_name).exists()
            ):
                shutil.copy2(self._paths.inp_entities / file_name, self._paths.outp_entities / file_name)


CRITERIA_MAPPING: dict[str, Callable[["Chunk", "Entity"], bool]] = {
    "inhabited_time<15s": lambda chunk, _: chunk.InhabitedTime <= 1200 * 0.25,
    "inhabited_time<30s": lambda chunk, _: chunk.InhabitedTime <= 1200 * 0.5,
    "inhabited_time<1m": lambda chunk, _: chunk.InhabitedTime <= 1200,
    "inhabited_time<2m": lambda chunk, _: chunk.InhabitedTime <= 1200 * 2,
    "inhabited_time<3m": lambda chunk, _: chunk.InhabitedTime <= 1200 * 3,
    "inhabited_time<5m": lambda chunk, _: chunk.InhabitedTime <= 1200 * 5,
    "inhabited_time<10m": lambda chunk, _: chunk.InhabitedTime <= 1200 * 10,
}


def process_region(manager: RegionManager, criteria: Callable[[Chunk, Entity], bool], file_name: str):
    region: Region = manager.open_file(file_name=file_name)
    manager.trim(region=region, condition=criteria)
    manager.save_to_file(region=region, file_name=file_name)


def process_batch(manager: RegionManager, criteria: str, file_names: list[str]) -> list[tuple[Exception, str]]:
    l = len(file_names)
    exceptions: list[tuple[Exception, str]] = []
    for i, r in enumerate(file_names, start=1):
        print(f"Processing region {r} ({i}/{l})")
        try:
            process_region(manager, CRITERIA_MAPPING[criteria], r)
        except AssertionError as e:
            e.add_note(f"[E]: AssertionError while processing {r}")
            tb = str(traceback.extract_tb(sys.exc_info()[2]))
            exceptions.append((e, tb))
        except Exception as e:
            e.add_note(f"[E]: Exception while processing {r}")
            tb = str(traceback.extract_tb(sys.exc_info()[2]))
            exceptions.append((e, tb))
    return exceptions


def main(*, threads: int | None, paths: Paths, trimming_criteria: str) -> None:
    rm = RegionManager(paths=paths)
    region_file_names: Iterable[str] = RegionLike.get_regions(paths.inp_region)

    if threads is None:
        res = process_batch(
            manager=rm,
            criteria=trimming_criteria,
            file_names=list(region_file_names),
        )
        for e, traceback in res:
            print("\n".join(e.__notes__), e, traceback)
    else:
        work: list[list[str]] = [[] for _ in range(threads)]
        for i, r in enumerate(region_file_names):
            work[i % threads].append(r)

        foo = partial(process_batch, rm, trimming_criteria)
        with Pool(threads) as p:
            res = p.map(func=foo, iterable=work)
            pass
        for combo in res:
            for e, traceback in combo:
                print("\n".join(e.__notes__), e, traceback)
